from .imports import *

def get_initial_caller() -> str:
    """
    Return the TRUE original caller: the entrypoint script that launched the program.
    """
    main_mod = sys.modules.get('__main__')
    
    # interactive environments (REPL) may have no __file__
    if not main_mod or not hasattr(main_mod, '__file__'):
        return None

    return os.path.realpath(main_mod.__file__)
def get_initial_caller_dir() -> str:
    """
    Return the directory of the TRUE original entrypoint script.
    """
    caller = get_initial_caller()
    return os.path.dirname(caller) if caller else None

def get_caller(i: Optional[int] = None) -> str:
    """
    Return the filename of the calling frame.

    Args:
        i: Optional stack depth offset. 
           None = immediate caller (depth 1).

    Returns:
        Absolute path of the file for the stack frame.
    """
    depth = 1 if i is None else int(i)
    stack = inspect.stack()
    if depth >= len(stack):
        depth = len(stack) - 1
    return stack[depth].filename


def get_caller_path(i: Optional[int] = None) -> str:
    """
    Return the absolute path of the caller's file.
    """
    depth = 1 if i is None else int(i)
    file_path = get_caller(depth + 1)
    return os.path.realpath(file_path)


def get_caller_dir(i: Optional[int] = None) -> str:
    """
    Return the absolute directory of the caller's file.
    """
    depth = 1 if i is None else int(i)
    abspath = get_caller_path(depth + 1)
    return os.path.dirname(abspath)



import inspect, importlib
from .context_utils import ensure_caller_package

def safe_import(
    name: str,
    *,
    member: str | None = None,
    package: str | None = None,
    file: str | None = None,
    caller_globals: dict | None = None,
):
    """
    Safe dynamic import that resolves relative imports when run as a script.
    """
    if file is None:
        frame = inspect.currentframe()
        assert frame is not None
        outer = frame.f_back
        caller_file = (outer.f_globals.get("__file__") if outer else None) or __file__
    else:
        caller_file = file

    if name.startswith(".") and not package:
        package = ensure_caller_package(caller_file, caller_globals=caller_globals)

    mod = importlib.import_module(name, package=package)
    return getattr(mod, member) if member else mod



from __future__ import annotations
from typing import Any, Iterable, Iterator, Tuple, Union, Dict, List, Optional
import re
from difflib import SequenceMatcher

JSONLike = Union[dict, list, tuple, set, str, int, float, bool, None]
PathType = Tuple[Union[str, int], ...]

def iter_values(obj: JSONLike, path: PathType = ()) -> Iterator[Tuple[PathType, Any]]:
    if isinstance(obj, dict):
        for k, v in obj.items():
            yield from iter_values(v, path + (k,))
    elif isinstance(obj, (list, tuple)):
        for i, v in enumerate(obj):
            yield from iter_values(v, path + (i,))
    elif isinstance(obj, set):
        for v in obj:
            yield from iter_values(v, path + ('<setitem>',))
    else:
        yield path, obj

def _norm(s: str, case_insensitive: bool) -> str:
    return s.lower() if case_insensitive else s

def _word_boundary_regex(term: str, case_insensitive: bool) -> re.Pattern:
    flags = re.IGNORECASE if case_insensitive else 0
    # word boundary around the whole term; escape term for literal
    return re.compile(rf"\b{re.escape(term)}\b", flags)

def _score_string(
    value: str,
    terms: Iterable[str],
    *,
    case_insensitive: bool = True,
    min_ratio: float = 0.6,
) -> Tuple[float, Dict[str, float]]:
    """
    Score a *string* against the list of terms. Returns (score, per_term_scores).
    Weights:
      exact=+3, word-boundary=+2, substring=+1, fuzzy=+0.5*ratio (if ratio>=min_ratio).
      +0.2 bonus per extra unique term matched (beyond the first).
    """
    if not isinstance(value, str):
        return 0.0, {}
    v = _norm(value, case_insensitive)
    per_term: Dict[str, float] = {}
    matched_terms = 0

    for term in terms:
        t = _norm(term, case_insensitive)
        term_score = 0.0

        if v == t:
            term_score = max(term_score, 3.0)

        # word boundary
        if _word_boundary_regex(term, case_insensitive).search(value):
            term_score = max(term_score, 2.0)

        # substring
        if t in v:
            term_score = max(term_score, 1.0)

        # fuzzy similarity
        ratio = SequenceMatcher(None, v, t).ratio()  # 0..1
        if ratio >= min_ratio:
            term_score = max(term_score, 0.5 * ratio)

        if term_score > 0:
            matched_terms += 1
            per_term[term] = term_score

    # Bonus for covering more than one term
    bonus = 0.2 * max(0, matched_terms - 1)
    total = sum(per_term.values()) + bonus
    return total, per_term

def best_match(
    obj: JSONLike,
    *,
    terms: Iterable[str],
    case_insensitive: bool = True,
    min_ratio: float = 0.6,
    key_weight: Dict[str, float] | None = None,  # e.g. {'window_title': 1.5}
    coerce_non_strings: bool = False,
) -> Optional[Dict[str, Any]]:
    """
    Return the single best leaf string:
    { 'value': str, 'path': PathType, 'score': float, 'per_term': {term:score} }
    """
    key_weight = key_weight or {}
    best: Optional[Dict[str, Any]] = None

    for path, val in iter_values(obj):
        s = val
        if not isinstance(s, str):
            if coerce_non_strings:
                s = str(val)
            else:
                continue

        score, per_term = _score_string(s, terms, case_insensitive=case_insensitive, min_ratio=min_ratio)
        if score <= 0:
            continue

        # Apply path-based weight if any path segment matches a key in key_weight
        weight = 1.0
        for seg in path:
            if isinstance(seg, str) and seg in key_weight:
                weight *= key_weight[seg]
        weighted = score * weight

        cand = {'value': s, 'path': path, 'score': weighted, 'per_term': per_term}
        if best is None or weighted > best['score']:
            best = cand
        # Optional tie-breakers (prefer more terms matched, then shorter value)
        elif best is not None and abs(weighted - best['score']) < 1e-9:
            if len(per_term) > len(best['per_term']):
                best = cand
            elif len(per_term) == len(best['per_term']) and len(s) < len(best['value']):
                best = cand

    return best

def top_k_matches(
    obj: JSONLike,
    *,
    terms: Iterable[str],
    k: int = 5,
    **kwargs
) -> List[Dict[str, Any]]:
    """Return top-k matches sorted by score desc."""
    items: List[Dict[str, Any]] = []
    for path, val in iter_values(obj):
        s = val if isinstance(val, str) else (str(val) if kwargs.get('coerce_non_strings') else None)
        if s is None:
            continue
        score, per_term = _score_string(s, terms,
                                        case_insensitive=kwargs.get('case_insensitive', True),
                                        min_ratio=kwargs.get('min_ratio', 0.6))
        if score <= 0:
            continue
        weight = 1.0
        for seg in path:
            if isinstance(seg, str) and kwargs.get('key_weight', {}).get(seg):
                weight *= kwargs['key_weight'][seg]
        items.append({'value': s, 'path': path, 'score': score * weight, 'per_term': per_term})

    items.sort(key=lambda d: d['score'], reverse=True)
    return items[:k]



from .imports import *
from .module_imports import *
# -------------------------
# Config dataclass
# -------------------------

@dataclass
class ScanConfig:
    allowed_exts: Set[str]
    exclude_exts: Set[str]
    allowed_types: Set[str]
    exclude_types: Set[str]
    allowed_dirs: List[str] = field(default_factory=list)
    exclude_dirs: List[str] = field(default_factory=list)
    allowed_patterns: List[str] = field(default_factory=list)
    exclude_patterns: List[str] = field(default_factory=list)

@dataclass
class SearchParams(ScanConfig):
    directories: List[str] = field(default_factory=list)
    add: bool = False
    recursive: bool = True
    strings: List[str] = field(default_factory=list)
    total_strings: bool = False
    parse_lines: bool = False
    spec_line: Union[bool, int] = False
    get_lines: bool = False

@dataclass
class AllParams(SearchParams):
    cfg = None
    allowed: Optional[Callable[[str], bool]] = None
    include_files: bool = True
    recursive: bool = True
# -------------------------
# Default sets
# -------------------------
DEFAULT_ALLOWED_EXTS: Set[str] = {
    ".py", ".pyw",                             # python
    ".js", ".jsx", ".ts", ".tsx", ".mjs",      # JS/TS
    ".html", ".htm", ".xml",                   # markup
    ".css", ".scss", ".sass", ".less",         # styles
    ".json", ".yaml", ".yml", ".toml", ".ini",  # configs
    ".cfg", ".md", ".markdown", ".rst",        # docs
    ".sh", ".bash", ".env",                    # scripts/env
    ".txt"                                     # plain text
}

DEFAULT_EXCLUDE_TYPES: Set[str] = {
    "image", "video", "audio", "presentation",
    "spreadsheet", "archive", "executable"
}

# never want these—even if they sneak into ALLOWED
_unallowed = set(get_media_exts(DEFAULT_EXCLUDE_TYPES)) | {
    ".bak", ".shp", ".cpg", ".dbf", ".shx", ".geojson",
    ".pyc", ".prj", ".sbn", ".sbx"
}
DEFAULT_EXCLUDE_EXTS = {e.split('.')[-1] for e in _unallowed if e not in DEFAULT_ALLOWED_EXTS}

DEFAULT_EXCLUDE_DIRS: Set[str] = {
    "node_modules", "old","__pycache__", "backups", "backup",
    "backs", "trash", "depriciated", "old", "__init__"
}

DEFAULT_EXCLUDE_PATTERNS: Set[str] = {
    "__init__*", "*.tmp", "*.log", "*.lock", "*.zip","*~"
}
REMOTE_RE = re.compile(r"^(?P<host>[^:\s]+@[^:\s]+):(?P<path>/.*)$")
AllowedPredicate = Optional[Callable[[str], bool]]
DEFAULT_EXCLUDE_FILE_PATTERNS=DEFAULT_EXCLUDE_PATTERNS
DEFAULT_ALLOWED_PATTERNS: List[str] = ["*"]
DEFAULT_ALLOWED_DIRS: List[str] = ["*"]
DEFAULT_ALLOWED_TYPES: List[str] = ["*"]
CANONICAL_MAP = {
    "directories": ["directory", "directories", "dir","dirs","directory","directories","d","dirname", "paths", "path","roots","root"],
    "files":["file","filepath","file_path","files","filepaths","file_paths","paths", "path","f"],
    "allowed_exts": ["allow_ext", "allowed_ext", "include_ext", "include_exts", "exts_allowed"],
    "exclude_exts": ["exclude_ext", "excluded_ext", "excluded_exts", "unallowed_ext", "unallowed_exts"],
    "allowed_types": ["allow_type", "allowed_type", "include_type", "include_types", "types_allowed"],
    "exclude_types": ["exclude_type", "excluded_type", "excluded_types", "unallowed_type", "unallowed_types"],
    "allowed_dirs": ["allow_dir", "allowed_dir", "include_dir", "include_dirs", "dirs_allowed"],
    "exclude_dirs": ["exclude_dir", "excluded_dir", "excluded_dirs", "unallowed_dir", "unallowed_dirs"],
    "allowed_patterns": ["allow_pattern", "allowed_pattern", "include_pattern", "include_patterns", "patterns_allowed"],
    "exclude_patterns": ["exclude_pattern", "excluded_pattern", "excluded_patterns", "unallowed_pattern", "unallowed_patterns"],
    "add":["add"],
    "recursive":["recursive"],
    "strings":["strings"],
    "total_strings":["total_strings"],
    "parse_lines":["parse_lines"],
    "spec_line":["spec_line"],
    "get_lines":["get_lines"]
}
DEFAULT_ALLOWED_EXCLUDE_MAP={
    "allowed_exts": {"default":DEFAULT_ALLOWED_EXTS,"type":type(DEFAULT_ALLOWED_EXTS),"canonical":CANONICAL_MAP.get("allowed_exts")},
    "exclude_exts": {"default":DEFAULT_EXCLUDE_EXTS,"type":type(DEFAULT_EXCLUDE_EXTS),"canonical":CANONICAL_MAP.get("exclude_exts")},
    "allowed_types": {"default":DEFAULT_ALLOWED_TYPES,"type":type(DEFAULT_ALLOWED_TYPES),"canonical":CANONICAL_MAP.get("allowed_types")},
    "exclude_types": {"default":DEFAULT_EXCLUDE_TYPES,"type":type(DEFAULT_EXCLUDE_TYPES),"canonical":CANONICAL_MAP.get("exclude_types")},
    "allowed_dirs": {"default":DEFAULT_ALLOWED_DIRS,"type":type(DEFAULT_ALLOWED_DIRS),"canonical":CANONICAL_MAP.get("allowed_dirs")},
    "exclude_dirs": {"default":DEFAULT_EXCLUDE_DIRS,"type":type(DEFAULT_EXCLUDE_DIRS),"canonical":CANONICAL_MAP.get("exclude_dirs")},
    "allowed_patterns": {"default":DEFAULT_ALLOWED_PATTERNS,"type":type(DEFAULT_ALLOWED_PATTERNS),"canonical":CANONICAL_MAP.get("allowed_patterns")},
    "exclude_patterns": {"default":DEFAULT_EXCLUDE_PATTERNS,"type":type(DEFAULT_EXCLUDE_PATTERNS),"canonical":CANONICAL_MAP.get("exclude_patterns")},
}
DEFAULT_CANONICAL_MAP={
    "directories":{"default":[],"type":list,"canonical":CANONICAL_MAP.get("directories")},
    "files":{"default":[],"type":list,"canonical":CANONICAL_MAP.get("files")},
    **DEFAULT_ALLOWED_EXCLUDE_MAP,
    "allowed":{"default":None,"type":bool,"canonical":CANONICAL_MAP.get("allowed")},
    "add":{"default":False,"type":bool,"canonical":CANONICAL_MAP.get("add")},
    "recursive":{"default":True,"type":bool,"canonical":CANONICAL_MAP.get("recursive")},
    "strings":{"default":None,"type":list,"canonical":CANONICAL_MAP.get("strings")},
    "total_strings":{"default":False,"type":bool,"canonical":CANONICAL_MAP.get("total_strings")},
    "parse_lines":{"default":False,"type":bool,"canonical":CANONICAL_MAP.get("parse_lines")},
    "spec_line":{"default":False,"type":bool,"canonical":CANONICAL_MAP.get("spec_line")},
    "get_lines":{"default":False,"type":bool,"canonical":CANONICAL_MAP.get("get_lines")},
}





from .ensure_utils import *

def get_globs(items,recursive: bool = True,allowed=None,cfg=None,**kwargs):
    glob_paths = []
    allowed = get_allowed_predicate(allowed=allowed,cfg=cfg,**kwargs)
    items = [item for item in make_list(items) if item]
    for item in items:
        pattern = os.path.join(item, "**/*")  # include all files recursively\n
        nuItems = glob.glob(pattern, recursive=recursive)
        if allowed:
            nuItems = [nuItem for nuItem in nuItems if nuItem and allowed(nuItem)]
        glob_paths += nuItems
    return glob_paths
def get_allowed_files(items,allowed=True,cfg=None,**kwargs):
    allowed = get_allowed_predicate(allowed=allowed,cfg=cfg,**kwargs)
    return [item for item in items if item and os.path.isfile(item) and allowed(item)]
def get_allowed_dirs(items,allowed=False,cfg=None,**kwargs):
    allowed = get_allowed_predicate(allowed=allowed,cfg=cfg,**kwargs)
    return [item for item in items if item and os.path.isdir(item) and allowed(item)]

def get_filtered_files(items,allowed=None,files = [],cfg=None,**kwargs):
    allowed = get_allowed_predicate(allowed=allowed,cfg=cfg,**kwargs)
    glob_paths = get_globs(items,allowed=allowed,cfg=cfg,**kwargs)
    return [glob_path for glob_path in glob_paths if glob_path and os.path.isfile(glob_path) and glob_path not in files and allowed(glob_path)]
def get_filtered_dirs(items,allowed=None,dirs = [],cfg=None,**kwargs):
    allowed = get_allowed_predicate(allowed=allowed,cfg=cfg,**kwargs)
    glob_paths = get_globs(items,allowed=allowed,cfg=cfg,**kwargs)
    return [glob_path for glob_path in glob_paths if glob_path and os.path.isdir(glob_path) and glob_path not in dirs and allowed(glob_path)]

def get_all_allowed_files(items,allowed=None,cfg=None,**kwargs):
    dirs = get_all_allowed_dirs(items,allowed=allowed,cfg=cfg,**kwargs)
    files = get_allowed_files(items,allowed=allowed,cfg=cfg,**kwargs)
    nu_files = []
    for directory in dirs:
        files += get_filtered_files(directory,allowed=allowed,files=files,cfg=cfg,**kwargs)
    return files
def get_all_allowed_dirs(items,allowed=None,cfg=None,**kwargs):
    allowed = get_allowed_predicate(allowed=allowed,cfg=cfg,**kwargs)
    dirs = get_allowed_dirs(items,allowed=allowed,cfg=cfg,**kwargs)
    nu_dirs=[]
    for directory in dirs:
        nu_dirs += get_filtered_dirs(directory,allowed=allowed,dirs=nu_dirs,cfg=cfg,**kwargs)
    return nu_dirs

